#define VERSION "0.083(Beta_EpocEmx_Release)"

#include <stdio.h>
#include <string.h>

#define RET_OK              0
#define RET_BADPARAMETER    1
#define RET_FILENOTFOUND    2
#define RET_NOMEM           3
#define RET_FILEFORMATERROR 4



#define COPYBUFFERSIZE 4096
#ifndef __BORLANDC__
 #define huge
#endif /* __BORLANDC__ */

void DisplayHelp(void)
{
 fprintf(stderr,"\nUsage: P2TAP InputFile.P OutputFile.TAP [switches]\n\n");
 fprintf(stderr,"switches: -k         Erase(kill) original InputFile.P\n");
 fprintf(stderr,"          -pFILE     Append prolog FILE at begin of OutputFile.TAP\n");
 fprintf(stderr,"          -nNUM      First Code block number in OutputFile.TAP\n");
 fprintf(stderr,"          -eADDRESS  Set entry address (not yet implemented) \n");
} /* DisplayHelp () */

void PrintAddress(unsigned long Val)
{
 printf("%5u(0x%04x) ",unsigned(Val&0xffff),unsigned(Val&0xffff) );
 if(Val > 0xffff)
  printf("\nValue is out of Z80 address space and be masked!!! \a");
} /* PrintAddress() */

void MessageFileEnd(char* filename)
{
 fprintf(stderr, "\nUnexpected end of file: %s\n\a",filename);
} /* MessageFileEnd() */

unsigned char CalculateCRC(unsigned char ActualCRC, unsigned char huge *Buffer, unsigned int Len)
{
 unsigned char CRC=ActualCRC;
 for (unsigned int n=0;n<Len;n++)
  CRC^=*(Buffer++);
 return(CRC);
} /* CalculateCRC() */

int WriteTAPHeaderBlock(FILE *dst,unsigned int StartAddr,unsigned int Length,char *Name)
{
 struct {
  unsigned int  FullBlockLen;
  unsigned char FlagByte;
  unsigned char FileType;
  unsigned char FileName[10];
  unsigned int  Length;
  unsigned int  StartAddress;
  unsigned int  SecondWord;
  unsigned char CRC;
 }TAPHeader;

 TAPHeader.FullBlockLen = 19;
 TAPHeader.FlagByte     = 0;
 TAPHeader.FileType     = 3;
 
 for (int n=0,i=0;n<10;n++)
 {
  TAPHeader.FileName[n] = (Name[i]!=0) ? (Name[i]) : ' ';
  if(Name[i]!=0)
   i++;
 }
 TAPHeader.Length       = Length;
 TAPHeader.StartAddress = StartAddr;
 TAPHeader.SecondWord   = 0x8000;
 TAPHeader.CRC = CalculateCRC(0, &(TAPHeader.FlagByte),12);
 TAPHeader.CRC = CalculateCRC(TAPHeader.CRC, (unsigned char*)&(TAPHeader.Length),2);
 TAPHeader.CRC = CalculateCRC(TAPHeader.CRC, (unsigned char*)&(TAPHeader.StartAddress),2);
 TAPHeader.CRC = CalculateCRC(TAPHeader.CRC, (unsigned char*)&(TAPHeader.SecondWord),2); 
 
 fwrite(&TAPHeader.FullBlockLen, 2, 1, dst);
 fwrite(&TAPHeader.FlagByte,     1, 1, dst); 
 fwrite(&TAPHeader.FileType,     1, 1, dst);
 fwrite(&TAPHeader.FileName,     1, 10, dst);
 fwrite(&TAPHeader.Length,       2, 1, dst);
 fwrite(&TAPHeader.StartAddress, 2, 1, dst);
 fwrite(&TAPHeader.SecondWord,   2, 1, dst);

 if( (fwrite(&TAPHeader.CRC,1,1,dst) ) != 1 )
  return(1); // Bad Write TAP header
 return(0); // TAP header write ok.
} /* WriteTAPHeaderBlock() */


int WriteTAPDataBlock(FILE *src,FILE *dst,unsigned int Length)
{
 unsigned int BlockLength=Length+2; // +2, flag byte and CRC byte
 unsigned char FlagCRC=255;
 unsigned int ReadBlockSize;
 unsigned char huge *CopyBuffer; // Musi byt Huge v Sumicich nefunguje !!!

 CopyBuffer = new(unsigned char[COPYBUFFERSIZE]);
 if (CopyBuffer==NULL)
  return (2); // Cannot allocate memory

 //Write BlockLength
 if (fwrite(&BlockLength,1,2,dst) != 2)
  return(1); // Cannot write to output file.TAP

 //Write Flag (255)
 if (fwrite(&FlagCRC,1,1,dst) != 1)
  return(1); // Cannot write to output file.TAP

 //Write Data block
 while (Length>0)
 {
  if (Length>COPYBUFFERSIZE)
   ReadBlockSize = COPYBUFFERSIZE;
  else
   ReadBlockSize = Length;

  if ( fread(CopyBuffer,1,ReadBlockSize,src) != ReadBlockSize )
  {
   delete(CopyBuffer); // Deallocate memory
   return(3);  // Unexpected end of input file.P
  }

  FlagCRC = CalculateCRC(FlagCRC,CopyBuffer,ReadBlockSize);

  if ( fwrite(CopyBuffer,1,ReadBlockSize,dst) != ReadBlockSize )
  {
   delete(CopyBuffer); // Deallocate memory
   return(1);  // Cannot write to output file.TAP
  }

  Length-=ReadBlockSize;
 } /* while */

 //Write CRC byte
 if (fwrite(&FlagCRC,1,1,dst) != 1)
 {
  delete(CopyBuffer);  // Deallocate memory
  return(1); // Cannot write to output file.TAP
 }

 delete(CopyBuffer);  // Deallocate memory
 return(0);
} /* WriteTAPDataBlock() */


int CreateTAPBlock(FILE *src,FILE *dst,unsigned int StartAddr,unsigned int Length,unsigned int BlockIndex)
{
 char Name[11];

 sprintf (Name,"CODE_%05u",BlockIndex);
 if( WriteTAPHeaderBlock(dst,StartAddr,Length,Name) )
  return(1); // Bad Write TAP header, So return.

 // Write data block and return result
 return( WriteTAPDataBlock(src,dst,Length) );
} /* CreateTAPBlock() */

int AppendFile( FILE* src, FILE *dst)
{
 unsigned int ReadBlockSize;
 // Buffer for copying Prolog file
 unsigned char huge *CopyBuffer; // Musi byt Huge u v Sumicich nefunguje !!!

 CopyBuffer = new (unsigned char[COPYBUFFERSIZE]);
 if (CopyBuffer==NULL)
  return (2); // Cannot allocate memory

 while( (ReadBlockSize=fread(CopyBuffer,1,COPYBUFFERSIZE,src)) != 0 )
 {
  printf(".");
  if( fwrite(CopyBuffer,1,ReadBlockSize,dst) != ReadBlockSize )
  {
   delete(CopyBuffer);  // Deallocate memory
   return(1); // Cannot write to output file.TAP
  }
 } /*while*/
 delete(CopyBuffer);  // Deallocate memory
 return(0);
} /* AppendFile() */

/* ************************************************************************ */
/* ************************************************************************ */
/* ************************************************************************ */
/* ************************************************************************ */

int main(int argn,char *argv[])
{
 FILE *src=NULL,*dst=NULL,*prolog=NULL;
 char PrologFileName[255];

 unsigned long TotalBlockLengthSum=0;
 unsigned int  BlockIndex=0;
 unsigned long BlockNum=0;

 unsigned char TmpChar;
 unsigned int  TmpInt;

 unsigned long BlockStartAddress;
 unsigned int  BlockLength;
 unsigned char ExtBlockHeader;
 unsigned char ExtBlockSegment;
 unsigned char ExtBlockGran;


 char KillSourceFileFlag=0; // Don't delete source file
 long EntryAddress=0x10000; // Mark as not used
 PrologFileName[0]='\0';    // No prolog file

 // Test input parameters
 printf("\nStarting P2TAP convertor ver%s by P0KE 2004.\n\n",VERSION);

 if( (argn<3)||(argn>7) )
 {
  DisplayHelp();
  return(RET_BADPARAMETER);
 }

 //Analyse switches
 for(int n=3;n<argn;n++)
 {
  if( (argv[n][0]!='-')||(strlen(argv[n])==0)||(strlen(argv[n])<2) )
  {
   fprintf(stderr, "Unknown switch: '%s'!\n\a",argv[n]);
   return(RET_BADPARAMETER);
  }
  switch (argv[n][1])
  {
   case 'k':
    KillSourceFileFlag=1;
   break;
   case 'p':
    strcpy(PrologFileName,&(argv[n][2]));
   break;
   case 'n':
    TmpInt=sscanf(&(argv[n][2]),"%u",&BlockIndex);
    if( TmpInt!=1 )
    {
     fprintf(stderr, "Bad number in switch: '%s'!\n\a",argv[n]);
     return(RET_BADPARAMETER);
    }
   break;
   case 'e':
   break;
   default:
    fprintf(stderr, "Unknown switch: '%s'!\n\a",argv[n]);
    return(RET_BADPARAMETER);
  } /* switch() */
 } /* for */


 // Open input file
 if ( (src = fopen(argv[1], "rb") )== NULL)
 {
  fprintf(stderr, "Cannot open input file: %s !\n\a",argv[1]);
  return(RET_FILENOTFOUND); // File not found
 }

 //Analyse InputFile.P  
 TmpInt = 0; // Bug if sizeof(TmpInt) >2 must set rest bytes to zero 
 fread(&TmpInt,2,1,src);
 if ( TmpInt != 0x1489) 
 {
  printf("Bug %x \n",TmpInt);
  fprintf(stderr, "Input file: %s is not .P format!\n\a",argv[1]);
  fclose(src);
  return(RET_FILEFORMATERROR); //FileFormat error
 }

 //Open output file
 if ( (dst = fopen(argv[2], "wb") )== NULL)
 {
  fprintf(stderr, "Cannot open output file: %s !\n\a",argv[2]);
  fclose(src);
  return(RET_FILENOTFOUND); //File not found
 }

 //Open prolog file
 if ( strlen(PrologFileName) )
 {
  if ( (prolog = fopen(PrologFileName, "rb") )== NULL)
  {
   fprintf(stderr, "Cannot open prolog file: %s !\n\a",argv[2]);
   fclose(src);
   fclose(dst);
   return(RET_FILENOTFOUND); //File not found
  }
  else //Copy prolog file at begin of dst file
  {
   printf("Appending prolog file: %s ",PrologFileName);
   if( (TmpInt=AppendFile(prolog,dst)) !=0 )
   {
    fclose(src);
    fclose(dst);
    fclose(prolog);

    switch(TmpInt)
    {
     case(1):      //Cannot write to output file for saving prolog file
      fprintf(stderr, "Cannot write to output file: %s !\n\a",argv[2]);
      return(RET_FILENOTFOUND);
     break;
     case(2):      //Cannot allocate memory for saving prolog file
      fprintf(stderr, "Cannot allocate memory!\n\a");
      return(RET_NOMEM);
     break;
    } /* switch() */
   }
   fclose(prolog); // Close prolog file
   printf(" OK\n\n");
  }
 }/* if */

 //Processing InputFile.P
 printf("block type block name start address length (byte) end address  \n");
 printf("---------------------------------------------------------------\n");

 while( fread(&TmpChar,1,1,src) )
 {
  switch(TmpChar & 0xFF) // sizeof bug
  {
  // 0x51 Z80 Code block mark standard
   case(0x51):
    printf("Zx80       CODE_%05u ",BlockIndex);
    //Read BlockStart Address ----------------------------------------------
    BlockStartAddress = 0;  // sizeof bug
    if( (fread(&BlockStartAddress,1,4,src) ) != 4 )
    {
     MessageFileEnd(argv[1]);
     break;
    }
    //Print and analyse BlockStartAddress
    PrintAddress(BlockStartAddress);
    BlockStartAddress&=0xffff;

    //Read BlockLength
    BlockLength = 0; // sizeof bug
    if( ( fread(&BlockLength,1,2,src) ) != 2)
    {
     MessageFileEnd(argv[1]);
     break;
    }
    //Print and analyse BlockLength
    PrintAddress(BlockLength);

    switch (CreateTAPBlock(src,dst,(unsigned int)(BlockStartAddress&0xffff),BlockLength,BlockIndex) )
    {
     case(0):
      PrintAddress(BlockStartAddress+BlockLength-1);
      TotalBlockLengthSum+=BlockLength;
      BlockNum++; BlockIndex++;
     break;
     case (1):
      printf ("TAPWriteError!!\a");
     break;
     case (2):
      printf ("NoMemory!!\a");
     break;
     case (3):
     default:
      MessageFileEnd(argv[1]);
     break;
    } /* switch */
    printf("\n");
   break; // 0x51 Z80 Code block mark standard ending ^^^^^^^^^^^^^^^^^^^^^^

  // 0x81, 0x51 Z80 Code block mark (extended)
   case(0x81):
    fread(&ExtBlockHeader ,1,1,src);
    fread(&ExtBlockSegment,1,1,src);
    if( (fread(&ExtBlockGran ,1,1,src) ) != 1 )
    {
     MessageFileEnd(argv[1]); //Staci testovat konec souboru na poslednim miste
     break;
    }
    if (ExtBlockHeader != 0x51)
    {
     fprintf(stderr,"\nUnsupported processor type:(%0x2X) (not Z80)\n\a",ExtBlockHeader);
     break;
    }
    printf("Zx80ext    CODE_%05u ",BlockIndex);
    //Read BlockStart Address ----------------------------------------------
    BlockStartAddress = 0;  // sizeof bug
    if( (fread(&BlockStartAddress,1,4,src) ) != 4 )
    {
     MessageFileEnd(argv[1]);
     break;
    }
    //Print and analyse BlockStartAddress
    PrintAddress(BlockStartAddress);
    BlockStartAddress&=0xffff;

    //Read BlockLength
    BlockLength = 0;  //sizeof bug
    if( ( fread(&BlockLength,1,2,src) ) != 2)
    {
     MessageFileEnd(argv[1]);
     break;
    }
    //Print and analyse BlockLength
    PrintAddress(BlockLength);

    switch (CreateTAPBlock(src,dst,(unsigned int)(BlockStartAddress&0xffff),BlockLength,BlockIndex) )
    {
     case(0):
      PrintAddress(BlockStartAddress+BlockLength-1);
      TotalBlockLengthSum+=BlockLength;
      BlockNum++; BlockIndex++;
     break;
     case (1):
      printf ("TAPWriteError!!\a");
     break;
     case (2):
      printf ("NoMemory!!\a");
     break;
     case (3):
     default:
      MessageFileEnd(argv[1]);
     break;
    } /* switch */
    printf("\n");
   break; //0x81, 0x51 Z80 Code block mark (extended) ^^^^^^^^^^^^^^^^^^


  // 0x80 Set entry point mark ---------------------------------------------
   case(0x80):
    printf("<entry point>         ");
    //Read code entry address
    EntryAddress = 0;  // sizeof bug
    if( (fread(&EntryAddress,1,4,src) ) != 4 )
     fprintf(stderr, "\nUnexpected end of file: %s\n\a",argv[1]);

    PrintAddress(EntryAddress);
    EntryAddress&=0xffff;
    printf("\n");
   break; //0x80 Set entry point mark ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

  // 0x00 Creator string mark ----------------------------------------------
   case(0x00):
    printf("\ncreator : ");
    while( fread(&TmpChar,1,1,src) )
     printf("%c",TmpChar);
    printf("\n");
   break; //0x00 Creator string mark ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

  // Default error section -------------------------------------------------
   default:
    fprintf(stderr, "\nInput file: %s is not .P format!\n\a",argv[1]);
    return(RET_FILEFORMATERROR);
   break; //Default error section  ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
  }/* switch() */
 } /* while() */

 // Print Statistic
 printf("\nAltogether %lu (0x%08lx) bytes in %lu CODE blocks\n",TotalBlockLengthSum,TotalBlockLengthSum,BlockNum);
 if (!BlockNum)
  BlockNum = 1; // Test dividing by zero in (BlockNum)
 printf("Average CODE block length: %.2f (0x%04lx) bytes\n",(float)TotalBlockLengthSum/BlockNum,TotalBlockLengthSum/BlockNum);
 // Close all files
 fclose(src);
 fclose(dst);

 // Remove source .P file
 if (KillSourceFileFlag)
  remove(argv[1]);
 return(RET_OK);
} /* main() */
